/*
# Game Raccoon Revision 0 Source Code
# Copyright (C) 2019 Mathew Carr

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License Version 3
# with Additional Terms as found in the file
# LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/** 
 * File:   main.c
 * Author: Mathew Carr
 *
 * GameRaccoon Mega Drive firmware.
 * Hey, it was almost called Hypertronic Superdrive
 * (or Megatronic Superdrive (Genetronic Superesis in America),
 * so be happy.
 * 
 * Created on                                  12 November  2019, 16:47
 * But then created anew on                    20 December  2019,  4:07
 * 
 * Step 1, compile ANYTHING, maybe get the lights blinking alternately.
 * 
 * Awesome Possum...'s title screen working    21 December  2019, 12:51
 * Good menu integration, directory cache      29 December  2019
 */
 
#include <xc.h>
#include <stdint.h>

#define _DI()		__asm__ volatile("disi #0x3FFF")
#define _EI()		__asm__ volatile("disi #0")

#include "ff.h"
#include "diskio.h"

/* Configuration bits to be written to reserved area. */
// CONFIG2
#pragma config POSCMOD = NONE           // Primary Oscillator Select (Primary oscillator disabled)
#pragma config OSCIOFNC = ON            // Primary Oscillator Output Function (OSC2/CLKO/RC15 functions as port I/O (RC15))
#pragma config FCKSM = CSECME           // Clock Switching and Monitor (Clock switching is enabled, Fail-Safe Clock Monitor is enabled)
#pragma config FNOSC = FRCPLL           // Oscillator Select (Fast RC Oscillator with PLL module (FRCPLL))
#pragma config IESO = OFF               // Internal External Switch Over Mode (IESO mode (Two-Speed Start-up) disabled)

// CONFIG1
#pragma config WDTPS = PS32768          // Watchdog Timer Postscaler (1:32,768)
#pragma config FWPSA = PR128            // WDT Prescaler (Prescaler ratio of 1:128)
#pragma config WINDIS = OFF             // Watchdog Timer Window (Windowed Watchdog Timer enabled; FWDTEN must be 1)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (Watchdog Timer is disabled)
#pragma config ICS = PGx1               // Comm Channel Select (Emulator/debugger uses EMUC1/EMUD1)
#pragma config GWRP = OFF               // General Code Segment Write Protect (Writes to program memory are allowed)
#pragma config GCP = OFF                // General Code Segment Code Protect (Code protection is disabled)
#pragma config JTAGEN = OFF             // JTAG Port Enable (JTAG port is disabled)

/**
 * Internal Fast RC oscillator is 8 MHz.
 * I'm going to PLL this up to 32MHz.
 */
#define FCY  32000000L

#define HIGH 1
#define LOW  0

#define ABS(_x_) (((_x_) >= 0) ? (_x_) : (-(_x_)))

#define BIT_N_IS_SET(_n_, _value_)       (!!((_value_) & (1  << (_n_))))

#ifndef NULL
#define NULL 0
#endif

/**
 * Copies uint8_ts from src to dst.
 * Copies a maximum of max_cells cells,
 * even if this would result in the null terminator
 * not being copied.
 * Therefore if you're copying into a uint8_t foobar[10]
 * you must only ever copy 9 cells into it and ensure 
 * that foobar[9] is always 0x00.
 */
void strncpy(uint8_t *dst, const uint8_t *src, uint16_t max_cells)
{
   if (max_cells == 0) return;
   
   uint8_t copied_char;
   
   do
   {
      copied_char = *dst++ = *src++;
   } while (--max_cells && copied_char);
}

/**
 * Returns the length of the null terminated 8-bit string pointed to by s,
 * but only scans a maximum of max_cells cells.
 * Length is exclusive of the terminator.
 * Returns max_cells if the string exceeds this length.
 */
uint16_t strnlen(const uint8_t *s, uint16_t max_cells)
{
   uint16_t count = 0;
   
   while (*s++ && max_cells--) count++;
      
   return count;
}

/* Fatfs seems to be happy with either a forward or backslash, so let's make a decision. */
#define PATH_SLASH_CHARACTER '/'
#define PATH_SLASH_STRING    "/"

/**
 * Given a string separated by / characters,
 * change the final / into a null terminator
 * to produce a string that is the parent of
 * the input.
 */
void strparent(uint8_t *fullpathname)
{
   uint8_t *slash = NULL;
   
   uint8_t *search_cursor = fullpathname;
   
   while (*search_cursor)
   {
      if (*search_cursor == PATH_SLASH_CHARACTER)
      {
         slash = search_cursor;
      }
      
      search_cursor++;
   }
   
   if (slash == NULL)
   {
      infinite_loop_error();
   }
   
   *slash = 0;
}

/**
 * Pin assignments:
 * 
 * TOP ROW: (as installed)
 *  
 * Pin 1 is in upper right, rightmost on top row.
 * 
 *  1: E5  = CART_D3
 *  2: E6  = CART_D4
 *  3: E7  = CART_D10
 *  4: G6  = SPI2CLK
 *  5: G7  = SPI2MISO
 *  6: G8  = SPI2MOSI
 *  7: !MCLR
 *  8: G9  = SPI2!SS
 *  9: VSS
 * 10: VDD
 * 11: B5  = CART_D11
 * 12: B4  = MCU_FLASH!WE
 * 13: B3  = MCU_!LEDBLUE
 * 14: B2  = MCU_!LEDGREEN
 * 15: PICKITCLOCK
 * 16: PICKITDATA
 * 
 * LEFT ROW: (as installed)
 * 
 * Pin 17 is in upper left, topmost on left side.
 *
 * 17: B6  = UNUSED (KEEP TRISB6 = 1)
 * 18: B7  = MCU_GAME!BOOT
 * 19: VDD
 * 20: VSS
 * 21: B8  = CART_A0
 * 22: B9  = CART_A1
 * 23: B10 = CART_A2
 * 24: B11 = CART_A3
 * 25: VSS
 * 26: VDD
 * 27: B12 = CART_A4
 * 28: B13 = CART_A5
 * 29: B14 = CART_A6
 * 30: B15 = CART_A7
 * 31: F4  = CART_A20
 * 32: F5  = CART_A19
 * 
 * BOTTOM ROW: (as installed)
 * 
 * Pin 33 is in lower left, leftmost on lower row.
 * 
 * 33: F3  = CART_A11
 * 34: F2  = CART_A9
 * 35: F6  = CART_A16
 * 36: G3  = CART_A8
 * 37: G2  = CART_A10
 * 38: VDD
 * 39: C12 = MCU_FLASH!READY
 * 40: C15 = CART_A15
 * 41: VSS
 * 42: D8  = CART_A18
 * 43: D9  = CART_A17
 * 44: D10 = CART_A12
 * 45: D11 = CART_A14
 * 46: D0  = CART_A13
 * 47: C13 = MCU_FLASH!FORCEOE
 * 48: C14 = REGISTERCLK
 * 
 * RIGHT ROW: (as installed)
 * 
 * 49: D1  = MCU_!68000
 * 50: D2  = MCU_REGISTER!OE
 * 51: D3  = CART_D7
 * 52: D4  = CART_D0
 * 53: D5  = CART_D8
 * 54: D6  = CART_D15
 * 55: D7  = CART_D6
 * 56: VCAP
 * 57: ENVREG
 * 58: F0  = CART_D14
 * 59: F1  = CART_D1
 * 60: E0  = CART_D13
 * 61: E1  = CART_D9
 * 62: E2  = CART_D12
 * 63: E3  = CART_D5
 * 64: E4  = CART_D2
 */

/**
 * DATA BUS GPIO ASSIGNMENTS:
 * 
 * 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
 * D6  F0  E0  E2  B5  E7  E1  D5  D3  D7  E3  E6  E5  E4  F1  D4
 */

#define DATA_BUS_TRIS_D15 _TRISD6
#define DATA_BUS_TRIS_D14 _TRISF0
#define DATA_BUS_TRIS_D13 _TRISE0
#define DATA_BUS_TRIS_D12 _TRISE2
#define DATA_BUS_TRIS_D11 _TRISB5
#define DATA_BUS_TRIS_D10 _TRISE7
#define DATA_BUS_TRIS_D9  _TRISE1
#define DATA_BUS_TRIS_D8  _TRISD5
#define DATA_BUS_TRIS_D7  _TRISD3
#define DATA_BUS_TRIS_D6  _TRISD7
#define DATA_BUS_TRIS_D5  _TRISE3
#define DATA_BUS_TRIS_D4  _TRISE6
#define DATA_BUS_TRIS_D3  _TRISE5
#define DATA_BUS_TRIS_D2  _TRISE4
#define DATA_BUS_TRIS_D1  _TRISF1
#define DATA_BUS_TRIS_D0  _TRISD4

#define DATA_BUS_LAT_D15   _LATD6
#define DATA_BUS_LAT_D14   _LATF0
#define DATA_BUS_LAT_D13   _LATE0
#define DATA_BUS_LAT_D12   _LATE2
#define DATA_BUS_LAT_D11   _LATB5
#define DATA_BUS_LAT_D10   _LATE7
#define DATA_BUS_LAT_D9    _LATE1
#define DATA_BUS_LAT_D8    _LATD5
#define DATA_BUS_LAT_D7    _LATD3
#define DATA_BUS_LAT_D6    _LATD7
#define DATA_BUS_LAT_D5    _LATE3
#define DATA_BUS_LAT_D4    _LATE6
#define DATA_BUS_LAT_D3    _LATE5
#define DATA_BUS_LAT_D2    _LATE4
#define DATA_BUS_LAT_D1    _LATF1
#define DATA_BUS_LAT_D0    _LATD4

#define DATA_BUS_PORT_D15    _RD6
#define DATA_BUS_PORT_D14    _RF0
#define DATA_BUS_PORT_D13    _RE0
#define DATA_BUS_PORT_D12    _RE2
#define DATA_BUS_PORT_D11    _RB5
#define DATA_BUS_PORT_D10    _RE7
#define DATA_BUS_PORT_D9     _RE1
#define DATA_BUS_PORT_D8     _RD5
#define DATA_BUS_PORT_D7     _RD3
#define DATA_BUS_PORT_D6     _RD7
#define DATA_BUS_PORT_D5     _RE3
#define DATA_BUS_PORT_D4     _RE6
#define DATA_BUS_PORT_D3     _RE5
#define DATA_BUS_PORT_D2     _RE4
#define DATA_BUS_PORT_D1     _RF1
#define DATA_BUS_PORT_D0     _RD4

void DATA_BUS_set_to_input()
{
   DATA_BUS_TRIS_D0  = 1;
   DATA_BUS_TRIS_D1  = 1;
   DATA_BUS_TRIS_D2  = 1;
   DATA_BUS_TRIS_D3  = 1;
   DATA_BUS_TRIS_D4  = 1;
   DATA_BUS_TRIS_D5  = 1;
   DATA_BUS_TRIS_D6  = 1;
   DATA_BUS_TRIS_D7  = 1;
   DATA_BUS_TRIS_D8  = 1;
   DATA_BUS_TRIS_D9  = 1;
   DATA_BUS_TRIS_D10 = 1;
   DATA_BUS_TRIS_D11 = 1;
   DATA_BUS_TRIS_D12 = 1;
   DATA_BUS_TRIS_D13 = 1;
   DATA_BUS_TRIS_D14 = 1;
   DATA_BUS_TRIS_D15 = 1;
}

void DATA_BUS_set_to_output()
{
   DATA_BUS_TRIS_D0  = 0;
   DATA_BUS_TRIS_D1  = 0;
   DATA_BUS_TRIS_D2  = 0;
   DATA_BUS_TRIS_D3  = 0;
   DATA_BUS_TRIS_D4  = 0;
   DATA_BUS_TRIS_D5  = 0;
   DATA_BUS_TRIS_D6  = 0;
   DATA_BUS_TRIS_D7  = 0;
   DATA_BUS_TRIS_D8  = 0;
   DATA_BUS_TRIS_D9  = 0;
   DATA_BUS_TRIS_D10 = 0;
   DATA_BUS_TRIS_D11 = 0;
   DATA_BUS_TRIS_D12 = 0;
   DATA_BUS_TRIS_D13 = 0;
   DATA_BUS_TRIS_D14 = 0;
   DATA_BUS_TRIS_D15 = 0;
}

void DATA_BUS_set_LAT_to_value(uint16_t value)
{
   if (BIT_N_IS_SET( 0, value)) { DATA_BUS_LAT_D0  = 1; } else { DATA_BUS_LAT_D0  = 0; }
   if (BIT_N_IS_SET( 1, value)) { DATA_BUS_LAT_D1  = 1; } else { DATA_BUS_LAT_D1  = 0; }
   if (BIT_N_IS_SET( 2, value)) { DATA_BUS_LAT_D2  = 1; } else { DATA_BUS_LAT_D2  = 0; }
   if (BIT_N_IS_SET( 3, value)) { DATA_BUS_LAT_D3  = 1; } else { DATA_BUS_LAT_D3  = 0; }
   if (BIT_N_IS_SET( 4, value)) { DATA_BUS_LAT_D4  = 1; } else { DATA_BUS_LAT_D4  = 0; }
   if (BIT_N_IS_SET( 5, value)) { DATA_BUS_LAT_D5  = 1; } else { DATA_BUS_LAT_D5  = 0; }
   if (BIT_N_IS_SET( 6, value)) { DATA_BUS_LAT_D6  = 1; } else { DATA_BUS_LAT_D6  = 0; }
   if (BIT_N_IS_SET( 7, value)) { DATA_BUS_LAT_D7  = 1; } else { DATA_BUS_LAT_D7  = 0; }
   if (BIT_N_IS_SET( 8, value)) { DATA_BUS_LAT_D8  = 1; } else { DATA_BUS_LAT_D8  = 0; }
   if (BIT_N_IS_SET( 9, value)) { DATA_BUS_LAT_D9  = 1; } else { DATA_BUS_LAT_D9  = 0; }
   if (BIT_N_IS_SET(10, value)) { DATA_BUS_LAT_D10 = 1; } else { DATA_BUS_LAT_D10 = 0; }
   if (BIT_N_IS_SET(11, value)) { DATA_BUS_LAT_D11 = 1; } else { DATA_BUS_LAT_D11 = 0; }
   if (BIT_N_IS_SET(12, value)) { DATA_BUS_LAT_D12 = 1; } else { DATA_BUS_LAT_D12 = 0; }
   if (BIT_N_IS_SET(13, value)) { DATA_BUS_LAT_D13 = 1; } else { DATA_BUS_LAT_D13 = 0; }
   if (BIT_N_IS_SET(14, value)) { DATA_BUS_LAT_D14 = 1; } else { DATA_BUS_LAT_D14 = 0; }
   if (BIT_N_IS_SET(15, value)) { DATA_BUS_LAT_D15 = 1; } else { DATA_BUS_LAT_D15 = 0; }
}

uint16_t DATA_BUS_read_PORT_value()
{
   uint16_t value = 0;
   
   if (DATA_BUS_PORT_D0)  { value |= (1u <<  0); }
   if (DATA_BUS_PORT_D1)  { value |= (1u <<  1); }
   if (DATA_BUS_PORT_D2)  { value |= (1u <<  2); }
   if (DATA_BUS_PORT_D3)  { value |= (1u <<  3); }
   if (DATA_BUS_PORT_D4)  { value |= (1u <<  4); }
   if (DATA_BUS_PORT_D5)  { value |= (1u <<  5); }
   if (DATA_BUS_PORT_D6)  { value |= (1u <<  6); }
   if (DATA_BUS_PORT_D7)  { value |= (1u <<  7); }
   if (DATA_BUS_PORT_D8)  { value |= (1u <<  8); }
   if (DATA_BUS_PORT_D9)  { value |= (1u <<  9); }
   if (DATA_BUS_PORT_D10) { value |= (1u << 10); }
   if (DATA_BUS_PORT_D11) { value |= (1u << 11); }
   if (DATA_BUS_PORT_D12) { value |= (1u << 12); }
   if (DATA_BUS_PORT_D13) { value |= (1u << 13); }
   if (DATA_BUS_PORT_D14) { value |= (1u << 14); }
   if (DATA_BUS_PORT_D15) { value |= (1u << 15); }
      
   return value;
}

/**
 * ADDRESS BUS GPIO ASSIGNMENTS:
 * 
 * ( !B/G=21 ) 20  19  18  17  16  15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
 * (    B7   ) F4  F5  D8  D9  F6  C15 D11 D0  D10 F3  G2  F2  G3  B15 B14 B13 B12 B11 B10 B9  B8
 */

#define ADDRESS_BUS_TRIS_A21          _TRISB7
#define ADDRESS_BUS_TRIS_A20          _TRISF4
#define ADDRESS_BUS_TRIS_A19          _TRISF5
#define ADDRESS_BUS_TRIS_A18          _TRISD8
#define ADDRESS_BUS_TRIS_A17          _TRISD9
#define ADDRESS_BUS_TRIS_A16          _TRISF6
#define ADDRESS_BUS_TRIS_A15          _TRISC15
#define ADDRESS_BUS_TRIS_A14          _TRISD11
#define ADDRESS_BUS_TRIS_A13          _TRISD0
#define ADDRESS_BUS_TRIS_A12          _TRISD10
#define ADDRESS_BUS_TRIS_A11          _TRISF3
#define ADDRESS_BUS_TRIS_A10          _TRISG2
#define ADDRESS_BUS_TRIS_A9           _TRISF2
#define ADDRESS_BUS_TRIS_A8           _TRISG3
#define ADDRESS_BUS_TRIS_A7           _TRISB15
#define ADDRESS_BUS_TRIS_A6           _TRISB14
#define ADDRESS_BUS_TRIS_A5           _TRISB13
#define ADDRESS_BUS_TRIS_A4           _TRISB12
#define ADDRESS_BUS_TRIS_A3           _TRISB11
#define ADDRESS_BUS_TRIS_A2           _TRISB10
#define ADDRESS_BUS_TRIS_A1           _TRISB9
#define ADDRESS_BUS_TRIS_A0           _TRISB8

#define ADDRESS_BUS_LAT_A21           _LATB7
#define ADDRESS_BUS_LAT_A20           _LATF4
#define ADDRESS_BUS_LAT_A19           _LATF5
#define ADDRESS_BUS_LAT_A18           _LATD8
#define ADDRESS_BUS_LAT_A17           _LATD9
#define ADDRESS_BUS_LAT_A16           _LATF6
#define ADDRESS_BUS_LAT_A15           _LATC15
#define ADDRESS_BUS_LAT_A14           _LATD11
#define ADDRESS_BUS_LAT_A13           _LATD0
#define ADDRESS_BUS_LAT_A12           _LATD10
#define ADDRESS_BUS_LAT_A11           _LATF3
#define ADDRESS_BUS_LAT_A10           _LATG2
#define ADDRESS_BUS_LAT_A9            _LATF2
#define ADDRESS_BUS_LAT_A8            _LATG3
#define ADDRESS_BUS_LAT_A7            _LATB15
#define ADDRESS_BUS_LAT_A6            _LATB14
#define ADDRESS_BUS_LAT_A5            _LATB13
#define ADDRESS_BUS_LAT_A4            _LATB12
#define ADDRESS_BUS_LAT_A3            _LATB11
#define ADDRESS_BUS_LAT_A2            _LATB10
#define ADDRESS_BUS_LAT_A1            _LATB9
#define ADDRESS_BUS_LAT_A0            _LATB8

void ADDRESS_BUS_set_to_input()
{
   ADDRESS_BUS_TRIS_A0   = 1;
   ADDRESS_BUS_TRIS_A1   = 1;
   ADDRESS_BUS_TRIS_A2   = 1;
   ADDRESS_BUS_TRIS_A3   = 1;
   ADDRESS_BUS_TRIS_A4   = 1;
   ADDRESS_BUS_TRIS_A5   = 1;
   ADDRESS_BUS_TRIS_A6   = 1;
   ADDRESS_BUS_TRIS_A7   = 1;
   ADDRESS_BUS_TRIS_A8   = 1;
   ADDRESS_BUS_TRIS_A9   = 1;
   ADDRESS_BUS_TRIS_A10  = 1;
   ADDRESS_BUS_TRIS_A11  = 1;
   ADDRESS_BUS_TRIS_A12  = 1;
   ADDRESS_BUS_TRIS_A13  = 1;
   ADDRESS_BUS_TRIS_A14  = 1;
   ADDRESS_BUS_TRIS_A15  = 1;
   ADDRESS_BUS_TRIS_A16  = 1;
   ADDRESS_BUS_TRIS_A17  = 1;
   ADDRESS_BUS_TRIS_A18  = 1;
   ADDRESS_BUS_TRIS_A19  = 1;
   ADDRESS_BUS_TRIS_A20  = 1;
}

void ADDRESS_BUS_set_to_output()
{
   ADDRESS_BUS_TRIS_A0   = 0;
   ADDRESS_BUS_TRIS_A1   = 0;
   ADDRESS_BUS_TRIS_A2   = 0;
   ADDRESS_BUS_TRIS_A3   = 0;
   ADDRESS_BUS_TRIS_A4   = 0;
   ADDRESS_BUS_TRIS_A5   = 0;
   ADDRESS_BUS_TRIS_A6   = 0;
   ADDRESS_BUS_TRIS_A7   = 0;
   ADDRESS_BUS_TRIS_A8   = 0;
   ADDRESS_BUS_TRIS_A9   = 0;
   ADDRESS_BUS_TRIS_A10  = 0;
   ADDRESS_BUS_TRIS_A11  = 0;
   ADDRESS_BUS_TRIS_A12  = 0;
   ADDRESS_BUS_TRIS_A13  = 0;
   ADDRESS_BUS_TRIS_A14  = 0;
   ADDRESS_BUS_TRIS_A15  = 0;
   ADDRESS_BUS_TRIS_A16  = 0;
   ADDRESS_BUS_TRIS_A17  = 0;
   ADDRESS_BUS_TRIS_A18  = 0;
   ADDRESS_BUS_TRIS_A19  = 0;
   ADDRESS_BUS_TRIS_A20  = 0;
}

/**
 * N.B. This includes BOOT/!GAME
 */
void ADDRESS_BUS_set_LAT_to_value(uint32_t value)
{   
   uint16_t value16_l = value;
   uint16_t value16_h = value >> 16;
   
   if (BIT_N_IS_SET( 0,      value16_l)) { ADDRESS_BUS_LAT_A0   = 1; } else { ADDRESS_BUS_LAT_A0   = 0; }
   if (BIT_N_IS_SET( 1,      value16_l)) { ADDRESS_BUS_LAT_A1   = 1; } else { ADDRESS_BUS_LAT_A1   = 0; }
   if (BIT_N_IS_SET( 2,      value16_l)) { ADDRESS_BUS_LAT_A2   = 1; } else { ADDRESS_BUS_LAT_A2   = 0; }
   if (BIT_N_IS_SET( 3,      value16_l)) { ADDRESS_BUS_LAT_A3   = 1; } else { ADDRESS_BUS_LAT_A3   = 0; }
   if (BIT_N_IS_SET( 4,      value16_l)) { ADDRESS_BUS_LAT_A4   = 1; } else { ADDRESS_BUS_LAT_A4   = 0; }
   if (BIT_N_IS_SET( 5,      value16_l)) { ADDRESS_BUS_LAT_A5   = 1; } else { ADDRESS_BUS_LAT_A5   = 0; }
   if (BIT_N_IS_SET( 6,      value16_l)) { ADDRESS_BUS_LAT_A6   = 1; } else { ADDRESS_BUS_LAT_A6   = 0; }
   if (BIT_N_IS_SET( 7,      value16_l)) { ADDRESS_BUS_LAT_A7   = 1; } else { ADDRESS_BUS_LAT_A7   = 0; }
   if (BIT_N_IS_SET( 8,      value16_l)) { ADDRESS_BUS_LAT_A8   = 1; } else { ADDRESS_BUS_LAT_A8   = 0; }
   if (BIT_N_IS_SET( 9,      value16_l)) { ADDRESS_BUS_LAT_A9   = 1; } else { ADDRESS_BUS_LAT_A9   = 0; }
   if (BIT_N_IS_SET(10,      value16_l)) { ADDRESS_BUS_LAT_A10  = 1; } else { ADDRESS_BUS_LAT_A10  = 0; }
   if (BIT_N_IS_SET(11,      value16_l)) { ADDRESS_BUS_LAT_A11  = 1; } else { ADDRESS_BUS_LAT_A11  = 0; }
   if (BIT_N_IS_SET(12,      value16_l)) { ADDRESS_BUS_LAT_A12  = 1; } else { ADDRESS_BUS_LAT_A12  = 0; }
   if (BIT_N_IS_SET(13,      value16_l)) { ADDRESS_BUS_LAT_A13  = 1; } else { ADDRESS_BUS_LAT_A13  = 0; }
   if (BIT_N_IS_SET(14,      value16_l)) { ADDRESS_BUS_LAT_A14  = 1; } else { ADDRESS_BUS_LAT_A14  = 0; }
   if (BIT_N_IS_SET(15,      value16_l)) { ADDRESS_BUS_LAT_A15  = 1; } else { ADDRESS_BUS_LAT_A15  = 0; }
   if (BIT_N_IS_SET(16 - 16, value16_h)) { ADDRESS_BUS_LAT_A16  = 1; } else { ADDRESS_BUS_LAT_A16  = 0; }
   if (BIT_N_IS_SET(17 - 16, value16_h)) { ADDRESS_BUS_LAT_A17  = 1; } else { ADDRESS_BUS_LAT_A17  = 0; }
   if (BIT_N_IS_SET(18 - 16, value16_h)) { ADDRESS_BUS_LAT_A18  = 1; } else { ADDRESS_BUS_LAT_A18  = 0; }
   if (BIT_N_IS_SET(19 - 16, value16_h)) { ADDRESS_BUS_LAT_A19  = 1; } else { ADDRESS_BUS_LAT_A19  = 0; }
   if (BIT_N_IS_SET(20 - 16, value16_h)) { ADDRESS_BUS_LAT_A20  = 1; } else { ADDRESS_BUS_LAT_A20  = 0; }
   if (BIT_N_IS_SET(21 - 16, value16_h)) { ADDRESS_BUS_LAT_A21  = 1; } else { ADDRESS_BUS_LAT_A21  = 0; }
}

/**
 * Control line pin assigments:
 * 
 * MCU_!LEDBLUE          = B3      OUTPUT
 * MCU_!LEDGREEN         = B2      OUTPUT
 * MCU_!68000            = D1      OUTPUT
 * MCU_FLASH!WE          = B4      OUTPUT
 * MCU_FLASH_GAME/!BOOT  = B7      OUTPUT
 * MCU_FLASH_READY/!BUSY = C12     INPUT
 * MCU_FLASH_FORCE!OE    = C13     OUTPUT
 * MCU_REGISTER!OE       = D2      OUTPUT
 * MCU_REGISTERCLK       = C14 CN0 INPUT
 */

#define TRIS_MCU_NEGLED_B            _TRISB3
#define TRIS_MCU_NEGLED_G            _TRISB2
#define TRIS_MCU_NEG68000            _TRISD1
#define TRIS_MCU_FLASH_NEGWE         _TRISB4
#define TRIS_MCU_FLASH_GAME_NEGBOOT  _TRISB7
#define TRIS_MCU_FLASH_READY_NEGBUSY _TRISC12
#define TRIS_MCU_FLASH_NEGFORCEOE    _TRISC13
#define TRIS_MCU_REGISTER_NEGOE      _TRISD2
#define TRIS_MCU_REGISTERCLK         _TRISC14

#define LAT_MCU_NEGLED_B             _LATB3
#define LAT_MCU_NEGLED_G             _LATB2
#define LAT_MCU_NEG68000             _LATD1
#define LAT_MCU_FLASH_NEGWE          _LATB4
#define LAT_MCU_FLASH_GAME_NEGBOOT   _LATB7
#define PORT_MCU_FLASH_READY_NEGBUSY _RC12
#define LAT_MCU_FLASH_NEGFORCEOE     _LATC13
#define LAT_MCU_REGISTER_NEGOE       _LATD2
#define PORT_MCU_REGISTERCLK         _RC14

#define TRIS_MCU_SD_MOSI             _TRISG8
#define TRIS_MCU_SD_MISO             _TRISG7
#define TRIS_MCU_SD_CLK              _TRISG6
#define TRIS_MCU_SD_NEGSS            _TRISG9

#define LED_ON  0
#define LED_OFF 1

#define MCU_NEG68000_MEGA_DRIVE_CONNECTED  0
#define MCU_NEG68000_MEGA_DRIVE_ISOLATED   1

/* Make sure everything is INPUT ONLY before connecting to 68000 bus. */
/* Do not call this primitive directly, call MCU_NEG68000_connect_bus_safely(). */
void MCU_NEG68000_connect_bus() { LAT_MCU_NEG68000 = MCU_NEG68000_MEGA_DRIVE_CONNECTED; }
void MCU_NEG68000_isolate_bus() { LAT_MCU_NEG68000 = MCU_NEG68000_MEGA_DRIVE_ISOLATED;  }

#define MCU_FLASH_NEGWE_WRITE_IDLE         1
#define MCU_FLASH_NEGWE_WRITE_BEGIN        0

void MCU_FLASH_NEGWE_write_disable() { LAT_MCU_FLASH_NEGWE = MCU_FLASH_NEGWE_WRITE_IDLE;  }
void MCU_FLASH_NEGWE_write_begin()   { LAT_MCU_FLASH_NEGWE = MCU_FLASH_NEGWE_WRITE_BEGIN; }

#define MCU_FLASH_GAME_NEGBOOT_SET_BOOT    0
#define MCU_FLASH_GAME_NEGBOOT_SET_GAME    1

void MCU_FLASH_GAME_NEGBOOT_set_to_boot() { LAT_MCU_FLASH_GAME_NEGBOOT = MCU_FLASH_GAME_NEGBOOT_SET_BOOT; }
void MCU_FLASH_GAME_NEGBOOT_set_to_game() { LAT_MCU_FLASH_GAME_NEGBOOT = MCU_FLASH_GAME_NEGBOOT_SET_GAME; }

#define MCU_FLASH_READY_NEGBUSY_IS_READY   1
#define MCU_FLASH_READY_NEGBUSY_IS_BUSY    0

int MCU_FLASH_READY_NEGBUSY_get_status() { return !!PORT_MCU_FLASH_READY_NEGBUSY; }

#define MCU_FLASH_NEGFORCEOE_OUTPUT_IDLE   1
#define MCU_FLASH_NEGFORCEOE_OUTPUT_ACTIVE 0

void MCU_FLASH_NEGFORCEOE_output_disable() { LAT_MCU_FLASH_NEGFORCEOE = MCU_FLASH_NEGFORCEOE_OUTPUT_IDLE;   }
void MCU_FLASH_NEGFORCEOE_output_enable()  { LAT_MCU_FLASH_NEGFORCEOE = MCU_FLASH_NEGFORCEOE_OUTPUT_ACTIVE; }

#define MCU_REGISTER_NEGOE_OUTPUT_IDLE     1
#define MCU_REGISTER_NEGOE_OUTPUT_ACTIVE   0

void MCU_REGISTER_NEGOE_output_disable() { LAT_MCU_REGISTER_NEGOE = MCU_REGISTER_NEGOE_OUTPUT_IDLE;   }
void MCU_REGISTER_NEGOE_output_enable()  { LAT_MCU_REGISTER_NEGOE = MCU_REGISTER_NEGOE_OUTPUT_ACTIVE; }

void LED_B_set_on()  { LAT_MCU_NEGLED_B = LED_ON;  }
void LED_B_set_off() { LAT_MCU_NEGLED_B = LED_OFF; }
void LED_G_set_on()  { LAT_MCU_NEGLED_G = LED_ON;  }
void LED_G_set_off() { LAT_MCU_NEGLED_G = LED_OFF; }

void LED_B_set_state(int s)
{
   if (s) { LED_B_set_on(); } else { LED_B_set_off(); }
}
void LED_G_set_state(int s)
{
   if (s) { LED_G_set_on(); } else { LED_G_set_off(); }
}

void MCU_NEG68000_connect_bus_safely()
{   
   DATA_BUS_set_to_input();
   ADDRESS_BUS_set_to_input();
   MCU_FLASH_NEGFORCEOE_output_disable();   
   MCU_FLASH_NEGWE_write_disable();
   MCU_REGISTER_NEGOE_output_disable();
   
   MCU_NEG68000_connect_bus();
}

/**
 * The directionality and output value of all the control lines
 * used by the PIC.
 * By default the bus is connected, and the PIC address and data
 * buses are set to input so the Mega Drive will be able to
 * transparently attempt to boot from the Boot parition of the Flash.
 */
void MCU_control_lines_set_directionality_and_default_state()
{
   MCU_NEG68000_connect_bus_safely();
   MCU_FLASH_GAME_NEGBOOT_set_to_boot();
   MCU_FLASH_NEGWE_write_disable();
   MCU_FLASH_NEGFORCEOE_output_disable();
   MCU_REGISTER_NEGOE_output_disable();

   LED_B_set_off();
   LED_G_set_off();
   
   TRIS_MCU_NEG68000            = 0;
   TRIS_MCU_FLASH_GAME_NEGBOOT  = 0;
   TRIS_MCU_FLASH_NEGWE         = 0;
   TRIS_MCU_FLASH_NEGFORCEOE    = 0;
   TRIS_MCU_REGISTER_NEGOE      = 0;
   
   TRIS_MCU_FLASH_READY_NEGBUSY = 1;
   TRIS_MCU_REGISTERCLK         = 1;
   
   TRIS_MCU_NEGLED_B            = 0;
   TRIS_MCU_NEGLED_G            = 0;
   
   TRIS_MCU_SD_MOSI             = 0;
   TRIS_MCU_SD_MISO             = 1;
   TRIS_MCU_SD_CLK              = 0;
   TRIS_MCU_SD_NEGSS            = 0;
   
   AD1PCFG = 0xFFFF; /* All digital, no analogue! */
   
   DATA_BUS_set_to_input();
   ADDRESS_BUS_set_to_input();
}

/* TODO: OH NO the Revision 0 of the Raccoon can't do this!! */
/* TODO: Install a 10k resistor to VCC3V3 on the MCU_FLASHREADY pin. It's open-drain! */
void MCU_FLASH_wait_until_ready()
{
   volatile int ready_value = 0;
   do {
      ready_value = MCU_FLASH_READY_NEGBUSY_get_status();
   } while (ready_value != MCU_FLASH_READY_NEGBUSY_IS_READY);
}

volatile int timer_spinlock = 0;

void __attribute__ ((interrupt, no_auto_psv))_T1Interrupt(void)
{
   IFS0bits.T1IF   = 0;     /* Clear interrupt flag for Timer1 event. */
   timer_spinlock  = 0;     /* Free main thread. */
   T1CONbits.TON   = 0;     /* Disable Timer1. */
}

/**
 * Spinlock for fcy_cycles FCY cycles using Timer1.
 * FCY is 16MHz.
 * 16000 is 1/1000 second.
 */
void timer1_wait(uint16_t fcy_cycles)
{
   T1CONbits.TON   = 0;
   TMR1            = 0;
   PR1             = fcy_cycles; 
   IFS0bits.T1IF   = 0;     /* Clear interrupt flag for Timer1 event. */
   IEC0bits.T1IE   = 1;     /* Enable Timer1 interrupts. */   
   T1CONbits.TON   = 1;     /* Enable Timer1. */
           
   timer_spinlock = 1;
   while (timer_spinlock);
   
   T1CONbits.TON   = 0;     /* Disable Timer1. */
}

/**
 * Spinlock for milliseconds milliseconds using Timer1.
 * FCY is 16MHz.
 * 16000 FCYs is 1/1000 second.
 */
void timer1_wait_milliseconds(uint16_t milliseconds)
{
   while (milliseconds--)
   {
      timer1_wait(16000);
   }
}

void infinite_loop()
{
   while (1)
   {
      volatile int a = 0;
      a++;
   }
}

void infinite_loop_error() /* Alternating. */
{
   int state = 0;
   
   while (1)
   {
      LED_B_set_state(state);
      state ^= 1;
      LED_G_set_state(state);
      timer1_wait_milliseconds(250);
   }
}

void infinite_loop_error_2() /* Green blue blue blue. */
{
   int state = 0;
   
   while (1)
   {
      LED_B_set_state(state & 1);
      LED_G_set_state(state == 1);
      timer1_wait_milliseconds(250);
      state++;
      state &= 7;
   }
}

void infinite_loop_error_3() /* Green blue blank blue green blank. */
{
   int state = 0;
   
   while (1)
   {
      LED_B_set_state(state == 0 || state == 5);
      LED_G_set_state(state == 1 || state == 4);
      timer1_wait_milliseconds(250);
      state++;
      state &= 7;
   }
}

void infinite_loop_pulse()
{
   int highcycle,lowcycle,state = 0, state2;
   
   while (1)
   {
      state +=    1;
      state &= 1023;
      
      if (state < 400)
      {
         state2 = state;
      }
      else if (state < (1024-400))
      {
         state2 = 400;
      }
      else
      {
         state2 = 1024-state;
      }
      if (state2 < 2) state2 = 2;
      
      highcycle = state2;
      lowcycle  = 400 - highcycle;
      
      highcycle += 3;
      lowcycle += 3;
      
      highcycle *= 5;
      lowcycle *= 5;
      
      for (int a = 0; a < 10; a++)
      {
         LED_B_set_off();
         LED_G_set_on();
         timer1_wait(highcycle); /* In 1/16000000 of a seconds. */
         LED_G_set_off();
         LED_B_set_on();
         timer1_wait(lowcycle);
      }
   }
}

/**
 * Appends the contents of the string src onto the end of the
 * string dst.
 * If the result would be more than max_cells long, including
 * terminator, explode.
 */
void strncat(uint8_t *dst, const uint8_t *src, uint16_t max_cells)
{
   uint16_t dstlen = strnlen(dst, max_cells);
   uint16_t srclen = strnlen(src, max_cells);
   
   if ((dstlen + srclen + 1) > max_cells)
   {
      infinite_loop_error();
   }
   
   strncpy(dst + dstlen, src, srclen+1);
}

void __attribute__ ((interrupt, no_auto_psv))_T2Interrupt(void)
{
   IFS0bits.T2IF   = 0;     /* Clear interrupt flag for Timer2 event. */
   disk_timerproc();
}

/**
 * 1000Hz timer as required by FatFs.
 */
void timer2_setup_for_fatfs()
{
   T2CONbits.TON   =    0;
   T2CONbits.TCKPS = 0b11;  /* 256 prescale. */
   TMR2            =    0;
   PR2             =   63;  /* 63 * 256 = 16128 cycles. 16000000 / 16128 = 992Hz */ 
   IFS0bits.T2IF   =    0;  /* Clear interrupt flag for Timer2 event. */
   IEC0bits.T2IE   =    1;  /* Enable Timer2 interrupts. */   
   T2CONbits.TON   =    1;  /* Enable Timer2. */
}

/**
 * Uses the bit-toggle polling method of determining if the Flash
 * is idle or busy.
 * At the end of this routine the address bus is
 * output and the data bus is input.
 * It doesn't matter whether you're in bypass mode or not.
 */
void MCU_FLASH_wait_bit_toggle_polling()
{
   int r1, r2;
   
   DATA_BUS_set_to_input();
   
   MCU_FLASH_NEGFORCEOE_output_disable();
   MCU_FLASH_NEGWE_write_disable();
   
   do 
   {
      MCU_FLASH_NEGFORCEOE_output_enable();
      r1 = DATA_BUS_PORT_D6;
      MCU_FLASH_NEGFORCEOE_output_disable();
      MCU_FLASH_NEGFORCEOE_output_enable();
      r2 = DATA_BUS_PORT_D6;
      MCU_FLASH_NEGFORCEOE_output_disable();
   } while (r1 != r2);
}

/**
 * This structure is used to convert between contiguous byte sequences
 * read from files or stored in the PIC's ROM, and Flash words written
 * to or read from the Flash PROM.
 * 
 * PIC is little endian.
 * Motorola 68000 is big endian.
 * What a mess!
 */
typedef struct _BIG_ENDIAN_FLASH_WORD
{
   union
   {
      struct
      {
         uint8_t second_byte; /* Named as they appear in a ROM file as an array of bytes. */
         uint8_t first_byte;
      };
      uint16_t uint16_value;
   };
} BIG_ENDIAN_FLASH_WORD;

/**
 * Swap a PIC little endian uint16_t into a 68000 big endian value to be written to flash.
 * We use this a lot to store/retrieve values from directory cache records.
 */
uint16_t uint16_t_byte_swap(uint16_t value)
{
   return (((value & 0xFF00u) >> 8)
         | ((value & 0x00FFu) << 8));
}

/**
 * Read length_in_bytes bytes into MCU ram dest from Flash
 * starting at flash[address_in_words] and ascending.
 * length_in_bytes must be a multiple of two.
 * address_in_words is in the full address space of the Flash
 * including both Boot and Game halves.
 * It doesn't matter whether you're in bypass mode or not.
 * Data bus will be set to input.
 */
void FLASH_read(uint32_t address_in_words, uint8_t *dest, uint32_t length_in_bytes)
{
   DATA_BUS_set_to_input();
   MCU_FLASH_NEGFORCEOE_output_disable();      
   
   BIG_ENDIAN_FLASH_WORD read_data;
      
   do
   {
      ADDRESS_BUS_set_LAT_to_value(address_in_words++);
      MCU_FLASH_NEGFORCEOE_output_enable();  
      read_data.uint16_value = DATA_BUS_read_PORT_value();
      MCU_FLASH_NEGFORCEOE_output_disable();      
      *dest++ = read_data.first_byte;
      *dest++ = read_data.second_byte;
   } while (length_in_bytes -= 2);
}

/**
 * Write to a new file filename on the SD card
 * the contents of the Flash memory at the given address
 * for so many bytes, must be a multiple of two.
 */
FRESULT FLASH_read_to_file(const uint8_t *filename, uint32_t address_in_words, uint32_t length_in_bytes)
{
   MCU_FLASH_NEGFORCEOE_output_disable();   
   MCU_FLASH_NEGWE_write_disable();   
   
#define FLASH_READ_TO_FILE_BUFFER_SIZE_BYTES 256   
   FIL      fhandle;
   uint8_t  read_buffer[FLASH_READ_TO_FILE_BUFFER_SIZE_BYTES];   
   
   FRESULT  open_success            = f_open(&fhandle, filename, FA_CREATE_ALWAYS | FA_WRITE);
   if (open_success != FR_OK) infinite_loop_error_2();
      
   FSIZE_t  bytes_remaining_to_read = length_in_bytes;   
   
   while (bytes_remaining_to_read)
   {            
      int bytes_to_read = FLASH_READ_TO_FILE_BUFFER_SIZE_BYTES;
      if (bytes_remaining_to_read < bytes_to_read)
      {
         bytes_to_read = bytes_remaining_to_read;            
      }

      LED_G_set_on();
      FLASH_read(address_in_words, read_buffer, bytes_to_read);
      LED_G_set_off();    

      bytes_remaining_to_read -= bytes_to_read;
      address_in_words        += bytes_to_read/2;

      LED_G_set_on();
      UINT bytes_written;
      FRESULT write_success = f_write(&fhandle, read_buffer, bytes_to_read, &bytes_written);
      if (write_success != FR_OK) infinite_loop_error_2();
      LED_G_set_off();
   };
   
   f_close(&fhandle);
         
   return FR_OK;
}

/**
 * This is an extremely powerful command!!
 * Performs a full address and data WE assert cycle.
 * Requires the address and data buses to be set to output
 * and the 68000 bus to be isolated.
 */
void FLASH_command(uint32_t address_in_words, uint16_t value)
{   
   ADDRESS_BUS_set_LAT_to_value(address_in_words);
   DATA_BUS_set_LAT_to_value(value);
   
   MCU_FLASH_NEGWE_write_begin();
   MCU_FLASH_NEGWE_write_disable();
}

/**
 * This puts the Flash chip the 'unlock bypass mode' state where commands can be fired more quickly.
 * Data bus will be set to output.
 */
void FLASH_unlock_bypass_mode_begin()
{
   DATA_BUS_set_to_output();
   FLASH_command(0x0555, 0xAA);
   FLASH_command(0x02AA, 0x55);
   FLASH_command(0x0555, 0x20);
}

/**
 * This returns the Flash back to read mode from bypass mode.
 * There's no real reason to do this though since all the
 * commands I want are available in bypass mode, and reading
 * is unaffected.
 * Data bus will be set to output.
 */
void FLASH_unlock_bypass_mode_end()
{
   DATA_BUS_set_to_output();
   FLASH_command(0x0000, 0x90);
   FLASH_command(0x0000, 0x00);
}

/**
 * Erase the 64 KByte / 32 kiloword sector at the given address.
 * address_in_words is in the full address space of the Flash
 * including both Boot and Game halves.
 * Data bus will be set to output.
 * Requires unlock bypass to be active.
 */
void FLASH_erase_sector_at_address_in_words(uint32_t address_in_words)
{      
   DATA_BUS_set_to_output();
      
   /**
    * Sectors are 32 kiloword big, which is
    *                                  0b1000000000000000 addressing units.
    * Therefore all pins BELOW A15 are ignored. A15 is the first one read.
    * 
    *                    BIT NUMBER
    *                    33222222222211111111110000000000
    *                    10987654321098765432109876543210 */
   address_in_words &= 0b00000000001111111000000000000000;
   
   /**
    *                               2211111111
    *                               1098765432
    * 
    * If we're writing to A21-A12 = 1111111xxx then on the model 03
    * chip that I have the uppermost sectors are smaller and need to be
    * erased at a finer granularity.
    * Bear in mind A21 is BOOT/!GAME.
    */
   if ((address_in_words == 0b00000000001111111000000000000000))
   {
      /* We have to erase all the mini sectors in the boot zone. */
      for (int sector = 0; sector < 8; sector++)
      {
         FLASH_command(          0x0000,  0x80);
         FLASH_command(address_in_words,  0x30);
      
         /* Increment by the size of a mini sector, which is an increase in pin A12. */
         address_in_words += 0b00000000000000000001000000000000;
                  
         MCU_FLASH_wait_bit_toggle_polling();
      }
   }
   else
   {
      /* We can erase the sector normally. */
      FLASH_command(          0x0000,  0x80);
      FLASH_command(address_in_words,  0x30);
      
      MCU_FLASH_wait_bit_toggle_polling();
   }   
}

/**
 * Writes one word to the Flash.
 * The unlock bypass mode must already be enabled.
 * The multiple-write buffer is not used.
 */
void FLASH_program_word(uint32_t address_in_words, uint16_t value)
{
   DATA_BUS_set_to_output();
   
   FLASH_command(          0x0000,  0xA0);
   FLASH_command(address_in_words, value);
   
   MCU_FLASH_wait_bit_toggle_polling();
}

/**
 * Writes up to 256 bytes to the Flash.
 * The unlock bypass mode must already be enabled.
 * The buffer is used.
 * length_in_bytes must be <= 256
 * Must be an even number of bytes of course.
 * Data bus will be set to output.
 */
void FLASH_program_words_with_buffer(uint32_t address_in_words, const uint8_t *source, int length_in_bytes)
{   
   DATA_BUS_set_to_output();
         
   /* Round up if an odd number was supplied. We warned you! */
   length_in_bytes++;
   length_in_bytes &= ~1;
      
   FLASH_command(address_in_words,  0x25);
   FLASH_command(address_in_words, (length_in_bytes >> 1) - 1);
   
   uint32_t address_in_words_cursor = address_in_words;
   
   do 
   {
      BIG_ENDIAN_FLASH_WORD big_endian_word;

      big_endian_word.first_byte  = *source++;
      big_endian_word.second_byte = *source++;

      FLASH_command(address_in_words_cursor++, big_endian_word.uint16_value);
   } while (length_in_bytes -= 2);
   
   FLASH_command(address_in_words,  0x29);
   
   MCU_FLASH_wait_bit_toggle_polling();
}

#define FLASH_SECTOR_SIZE_BYTES                                                                                            (65536l)
#define FLASH_SECTOR_SIZE_WORDS                                                                         (FLASH_SECTOR_SIZE_BYTES/2)
#define FLASH_PARTITION_SIZE_BYTES                                                                                 (4l*1024l*1024l)
#define FLASH_PARTITION_SIZE_WORDS                                                                  (FLASH_PARTITION_SIZE_BYTES/2l)

#define FLASH_PARTITION_BOOT 0
#define FLASH_PARTITION_GAME 1

#define FLASH_PARTITION_BOOT_ADDRESS_WORDS                          0
#define FLASH_PARTITION_GAME_ADDRESS_WORDS FLASH_PARTITION_SIZE_WORDS

/**
 * Copies the full contents of the file filename to the Flash memory.
 * You can choose whether to overwrite the boot or game partitions,
 * you lucky duck!
 * The writing starts from the beginning of the partition for
 * as long as the file is.
 * If verify is non-zero, the lights will blink simultaneously
 * an error is found.
 */
FRESULT copy_sd_file_to_flash(const char *filename, unsigned int partition, unsigned int verify)
{     
#define COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES                                                                          (256)
#define COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_WORDS                               (COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES/2)
   
#define COPY_SD_FILE_TO_FLASH_READ_BUFFERS_PER_FLASH_SECTOR (FLASH_SECTOR_SIZE_BYTES/COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES)
      
   MCU_FLASH_NEGFORCEOE_output_disable();   
   MCU_FLASH_NEGWE_write_disable();   
   
   FIL      fhandle;
   UINT     no_bytes_read;   
   uint8_t  read_buffer[COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES];   
   uint8_t  verify_buffer[COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES];   
   
   FRESULT  open_success            = f_open(&fhandle, filename, FA_READ);
   if (open_success != FR_OK) infinite_loop_error_2();
   
   FSIZE_t  file_size               = f_size(&fhandle);   
   FSIZE_t  bytes_remaining_to_read = file_size;   
   
   uint32_t flash_address_in_words  = FLASH_PARTITION_BOOT_ADDRESS_WORDS;
   if (partition == FLASH_PARTITION_BOOT)
   {
      flash_address_in_words = FLASH_PARTITION_BOOT_ADDRESS_WORDS;
   }
   else
   {
      flash_address_in_words = FLASH_PARTITION_GAME_ADDRESS_WORDS;
   }
   
   int buffer_reads_before_next_sector_erase = 0;
            
   while (bytes_remaining_to_read)
   {
      if (buffer_reads_before_next_sector_erase == 0)
      {
         LED_G_set_on();
         FLASH_erase_sector_at_address_in_words(flash_address_in_words);
         LED_G_set_off();
         buffer_reads_before_next_sector_erase = COPY_SD_FILE_TO_FLASH_READ_BUFFERS_PER_FLASH_SECTOR;   
      }
      buffer_reads_before_next_sector_erase--;
      
      int bytes_to_read = COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES;
      if (bytes_remaining_to_read < bytes_to_read)
      {
         bytes_to_read = bytes_remaining_to_read;            
      }

      LED_B_set_on();
      FRESULT readsuccess = f_read(&fhandle, read_buffer, bytes_to_read, &no_bytes_read);        
      LED_B_set_off();    
      if (readsuccess != FR_OK) infinite_loop_error_2();

      bytes_remaining_to_read -= bytes_to_read;

      LED_G_set_on();
      FLASH_program_words_with_buffer(flash_address_in_words, read_buffer, COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES);
      LED_G_set_off();

      if (verify)
      {   
         FLASH_read(flash_address_in_words, verify_buffer, COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES);

         for (int byte_compare = 0; byte_compare < COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_BYTES; byte_compare++)
         {
            if (read_buffer[byte_compare] != verify_buffer[byte_compare])
            {               
               /* This is really bad!! How did this happen!! */
               infinite_loop_error_3();
            }
         }
      } 

      flash_address_in_words += COPY_SD_FILE_TO_FLASH_READ_BUFFER_SIZE_WORDS;
   };
   
   f_close(&fhandle);
         
   return FR_OK;
}

#define FLASH_DEBUG_READ_SIZE (1024*5)

void FLASH_debug_read(uint32_t address_in_words)
{
   uint8_t readflash[FLASH_DEBUG_READ_SIZE];
   
   for (int x = 0; x < FLASH_DEBUG_READ_SIZE; x++)
   {
      readflash[x] = x;
   }
   
   FLASH_read(address_in_words, readflash, FLASH_DEBUG_READ_SIZE);
      
   volatile int a = 0;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
   a++;
}

/**
 * CN is used to detect changes to REGISTERCLK when a successful
 * !TIME write takes place.
 */
volatile int CN0_event_taken_place = 0;

void __attribute__ ((interrupt, no_auto_psv))_CNInterrupt(void)
{
   IEC1bits.CNIE   = 0; /* Disable interrupts. */
   IFS1bits.CNIF   = 0; /* Clear interrupt flag for CN event. */
   CN0_event_taken_place = 1;
}

/**
 * Put the Game Raccoon into an idle state waiting for a 
 * !TIME write to the on-board register.
 * The bus will be connected during this waiting time.
 * GR will take no action until a !TIME write takes place.
 * When this takes place, the GR will wake up after a 0.1second delay
 * and isolate the bus and read the contained value.
 * The 16-bit value written to !TIME by the Mega Drive is returned.
 * At the end of this function, the 68000 bus will be isolated,
 * the address bus set to output and the data bus set to input.
 */
uint16_t MCU_REGISTERCLK_idle_for_changes()
{   
   MCU_NEG68000_connect_bus_safely(); /* Make sure everything is INPUT ONLY before executing this. */
   
   /* Enable interrupts for CN on CN0 (C14 / REGISTERCLK) */
   CN0_event_taken_place = 0;
   
   CNEN1bits.CN0IE = 1; /* Watch CN0 / C14 */
   IFS1bits.CNIF   = 0; /* Clear interrupt flag. */
   IEC1bits.CNIE   = 1; /* Enable interrupts. */
      
   while (!CN0_event_taken_place) {}
   
   /* Wait for 0.10 second for the Mega Drive to definitively jump to executing from RAM. */
   timer1_wait_milliseconds(10);
   
   MCU_NEG68000_isolate_bus();
   MCU_REGISTER_NEGOE_output_enable();
   uint16_t read_value = DATA_BUS_read_PORT_value();
   MCU_REGISTER_NEGOE_output_disable();
   
   ADDRESS_BUS_set_to_output();
   
   return read_value;
}

#define MAKE_16_BIT_VALUE_FROM_CHARS(_msb_,_lsb_) ((uint16_t)((((uint16_t)(_msb_)) << 8) | ((uint16_t)(_lsb_))))

#define REGISTERCLK_CONSTANT_GR_HANDSHAKE               (MAKE_16_BIT_VALUE_FROM_CHARS('G','R'))
#define REGISTERCLK_CONSTANT_BOOT_INSTALLED_GAME        0xFFFFu
#define REGISTERCLK_CONSTANT_REGENERATE_DIRECTORY_CACHE 0xFFFEu

/**
 * Hang and wait for the first !TIME write to take place.
 * If the first word received is GR, then return still awake
 * and isolated from the bus.
 * Otherwise, enter an infinite loop with the bus isolated.
 * (The Mega Drive will be suspended indefinitely.)
 */
void game_raccoon_wait_for_handshake()
{
   /**
    * If the handshake fails (if the first !TIME write is not
    * GR or it's read incorrectly), loop infinitely.
    */
   uint16_t register_value_handshake = MCU_REGISTERCLK_idle_for_changes();
   if (register_value_handshake != REGISTERCLK_CONSTANT_GR_HANDSHAKE)
   {   
      infinite_loop();
   }
}

/**
 * We've received the 0xFFFF command to boot the installed game.
 * We need to disable the PIC entirely, make it inert and invisible.
 * Setting the NEGBOOT/GAME flag to game presents the game partition
 * to the Mega Drive.
 */
void game_raccoon_switch_to_game_mode_and_terminate()
{  
   MCU_FLASH_GAME_NEGBOOT_set_to_game();
   
   LED_B_set_off(); /* Setting these to flash on a timer sure would be distracting... */
   LED_G_set_off();
         
   MCU_NEG68000_connect_bus_safely();   

   infinite_loop();
}

/**
 * Used by fatfs writing functions for timestamps.
 * Force it to zero Unix Epoch time and ignore it since we never write.
 */
DWORD get_fattime(void)
{
   return 0;
}

/**
 * Global structure used for holding state about file operations.
 */ 
FATFS fatfs;

/**
 * Filenames are stored as full absolute paths, zero terminated.
 * Filenames are stored first in the record regardless of record
 * type so the 68000 side can easily sort them.
 */
typedef struct _DIRECTORY_CACHE_RECORD
{
   union
   {
#define DIRECTORY_CACHE_RECORD_FILE_FULLPATHNAME_FIELD_LENGTH (256)
#define DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH (256 - (2*4)) /* Four fields. */
      struct
      {
         uint8_t  fullpathname[DIRECTORY_CACHE_RECORD_FILE_FULLPATHNAME_FIELD_LENGTH];
      } file_record;
      struct
      {
         uint8_t  fullpathname[DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH];
         uint16_t directory_count;
         uint16_t file_count;
         uint16_t record_index_contents_begin;
         uint16_t record_index_parent;
      } directory_record;
   };
} DIRECTORY_CACHE_RECORD;

#define DIRECTORY_CACHE_RECORD_SIZE_BYTES (256l) /* Defined here. I've made it the size of a Flash write buffer. */
#define DIRECTORY_CACHE_RECORD_SIZE_WORDS (DIRECTORY_CACHE_RECORD_SIZE_BYTES/2l)

#define GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_BYTES       0x200000ul
#define GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_WORDS       (GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_BYTES/2l)
#define GAME_RACCOON_DIRECTORY_CACHE_ENTRIES_CAPACITY             1024ul
#define GAME_RACCOON_DIRECTORY_CACHE_END_ADDRESS_BYTES    (GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_BYTES + (sizeof(DIRECTORY_CACHE_RECORD) * GAME_RACCOON_DIRECTORY_CACHE_ENTRIES_CAPACITY))

#define DIRECTORY_CACHE_RECORD_COUNT_PER_FLASH_SECTOR (FLASH_SECTOR_SIZE_WORDS/DIRECTORY_CACHE_RECORD_SIZE_WORDS) 

uint32_t DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_bytes(uint32_t record_number)
{
   return GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_BYTES + (DIRECTORY_CACHE_RECORD_SIZE_BYTES * record_number);
}
uint32_t DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(uint16_t record_number)
{
   return DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_bytes(record_number)/2l;
}

/**
 * Fields we do not want to have pushed onto the stack as
 * the generation algorithm runs recursively.
 */
typedef struct _GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE
{   
   /* Name of the current path that is being scanned for files and folders. */
   char                   present_scanned_directory[256];
   
   DIR                    dirhandle;
   FILINFO                fileinfo;
   
   uint16_t               record_index_write_cursor;
} GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE;

#define DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_DIRECTORY_COUNT             0
#define DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_FILE_COUNT                  1
#define DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_CONTENTS_BEGIN 2
#define DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_PARENT         3
#define DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_MAX                 DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_PARENT

/**
 * Writes to the Flash the given field for the given record as the given value.
 * You better have erased the Flash sector that this corresponds to first!
 */
void FLASH_game_raccoon_directory_cache_program_record_field(uint16_t record_index, uint16_t field, uint16_t value)
{
   if (field > DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_MAX) infinite_loop_error();
     
   uint32_t write_offset_into_record_bytes = DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH + (field * 2);
   uint32_t write_offset_into_record_words = write_offset_into_record_bytes/2;
   
   uint32_t address_of_record_words = DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(record_index);
   
   uint32_t destination_program_address_words = address_of_record_words + write_offset_into_record_words;
   
   FLASH_program_word(destination_program_address_words, value);
}

/**
 * All types of records begin with a filename.
 * Use this function to program the filename part of a record.
 */
void FLASH_game_raccoon_directory_cache_program_record_filename(uint16_t record_index, const uint8_t *s)
{   
   uint32_t destination_program_address_words = DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(record_index);
   
   /* Abort if the string we're trying to write won't fit in the filename field. */
   uint32_t length = strnlen(s, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);   
   if (length >= (DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH-1))
   {
      infinite_loop_error();
   }
   
   FLASH_program_words_with_buffer(destination_program_address_words, s, length+1);
}

void GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE_increment_record_index_write_cursor_erase_sector_if_necessary(
   GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE *w
)
{
   w->record_index_write_cursor++;
   
   if (w->record_index_write_cursor >= GAME_RACCOON_DIRECTORY_CACHE_ENTRIES_CAPACITY)
   {
      /* Too many records! */
      infinite_loop_error();
   }
   
   if ((w->record_index_write_cursor % DIRECTORY_CACHE_RECORD_COUNT_PER_FLASH_SECTOR) == 0)
   {
      /* The next record is on a new sector, we need to erase that sector. */
      LED_G_set_on();
      FLASH_erase_sector_at_address_in_words(DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(w->record_index_write_cursor));
      LED_G_set_off();
   } 
}

/**
 * Iterate through the contents of the current directory
 * (which is determined by w->present_scanned_directory)
 * and generate the list of directory and file records
 * for it.
 * The currently scanning directory has a record
 * at index scanning_directory_record_index.
 * After each of its children records have been generated,
 * enter each of the directory records that have been generated
 * and populate them.
 * 
 * Going into this function, the record scanning_directory_record_index
 * is populated with the index of its parents and the index of the
 * start of its own listing of contents, but its directory and file counts
 * are still unprogrammed.
 */
void game_raccoon_regenerate_directory_cache_populate_record(
                                            uint16_t  scanning_directory_record_index,
   GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE *w
)
{   
   /* Load the pathname of the directory we're entering into w->present_scanned_directory. */
   FLASH_read(DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(scanning_directory_record_index), w->present_scanned_directory, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
      
   /**
    * Before I write anything, if I contain any directories, they will
    * be found at the following record index:
    */
   uint16_t my_directories_list_record_index = w->record_index_write_cursor+1;
   
   /**
    * Our contents if we have any will appear at the
    * current writing cursor's position (after it has
    * advanced).
    * 
    * If we have no contents, then that's fine.
    * We don't advance the cursor here, only when
    * an entry has been located.
    */
   LED_B_set_on();
   FLASH_game_raccoon_directory_cache_program_record_field(scanning_directory_record_index, DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_CONTENTS_BEGIN, my_directories_list_record_index);
   LED_B_set_off();
      
   uint16_t directories_count = 0;
   uint16_t files_count       = 0;
   
   
   if (f_opendir(&(w->dirhandle), w->present_scanned_directory) != FR_OK)
   {
      infinite_loop_error();
   }
   
   while (1)
   {      
      if (f_readdir(&(w->dirhandle), &(w->fileinfo)) != FR_OK)
      {
         infinite_loop_error();
      }
      
      if (w->fileinfo.fname[0])
      {
         /**
          * An entry in the current SD card directory has been located.
          * Determine if it's a file or a directory.
          * We want to output directories first, so ignore all files on this pass.
          */
         if (!(w->fileinfo.fattrib & AM_DIR)) continue;
         
         /**
          * Allocate a new record for this directory and write
          * its fullpathname "./OLD/PATH/newpath" into the filename field.
          * Also write the parent record index field.
          * The other fields are written when we enter that directory later.
          */
         GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE_increment_record_index_write_cursor_erase_sector_if_necessary(w);
         
         strncat(w->present_scanned_directory, PATH_SLASH_STRING, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
         strncat(w->present_scanned_directory, w->fileinfo.fname, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
         LED_B_set_on();
         FLASH_game_raccoon_directory_cache_program_record_filename(w->record_index_write_cursor, w->present_scanned_directory);
         FLASH_game_raccoon_directory_cache_program_record_field(w->record_index_write_cursor, DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_PARENT, scanning_directory_record_index);
         LED_B_set_off();
         /* From ./OLD/PATH/newpath back to ./OLD/PATH */
         strparent(w->present_scanned_directory);
         
         directories_count++;
      }
      else
      {
         /* No more entries in this directory: the directory gathering is complete. */
         f_closedir(&(w->dirhandle));
         break;
      }
   }
   
   if (f_opendir(&(w->dirhandle), w->present_scanned_directory) != FR_OK)
   {
      infinite_loop_error();
   }
   
   while (1)
   {      
      if (f_readdir(&(w->dirhandle), &(w->fileinfo)) != FR_OK)
      {
         infinite_loop_error();
      }
      
      if (w->fileinfo.fname[0])
      {
         /**
          * An entry in the current SD card directory has been located.
          * Determine if it's a file or a directory.
          * We want to output files second, so ignore all directories on this pass.
          */
         if (w->fileinfo.fattrib & AM_DIR) continue;
         
         /**
          * Allocate a new record for this file and write
          * its fullpathname "./OLD/PATH/filename.stuff" into the filename field.
          * There are no other fields to write.
          */
         GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE_increment_record_index_write_cursor_erase_sector_if_necessary(w);
         
         strncat(w->present_scanned_directory, PATH_SLASH_STRING, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
         strncat(w->present_scanned_directory, w->fileinfo.fname, DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
         LED_B_set_on();
         FLASH_game_raccoon_directory_cache_program_record_filename(w->record_index_write_cursor, w->present_scanned_directory);
         LED_B_set_off();
         /* From ./OLD/PATH/newpath back to ./OLD/PATH */
         strparent(w->present_scanned_directory);
         
         files_count++;
      }
      else
      {
         /* No more entries in this directory: the file gathering is complete. */
         f_closedir(&(w->dirhandle));
         break;
      }
   }
   
   /**
    * At this point:
    * w->record_index_write_cursor is the index of the last file item.
    * We have ascertained the number of directories and files in this
    * currently scanned directory.
    * Program them.
    */
   LED_B_set_on();
   FLASH_game_raccoon_directory_cache_program_record_field(scanning_directory_record_index, DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_DIRECTORY_COUNT, directories_count);
   FLASH_game_raccoon_directory_cache_program_record_field(scanning_directory_record_index, DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_FILE_COUNT,      files_count);
   LED_B_set_off();
      
   /**
    * We need to go through all of the directories that this
    * currently scanned directory has identified, and generate
    * their own lists.
    */
   for (uint16_t directory_index = 0; directory_index < directories_count; directory_index++)
   {
      game_raccoon_regenerate_directory_cache_populate_record(my_directories_list_record_index + directory_index, w);
   }
}

/**
 * Erase and regenerate the directory cache found in the Boot partition
 * at GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_WORDS.
 * The first record (record zero) is always the root of the SD card.
 * Records are always 256 bytes big.
 * Each record is either a Directory or a File.
 */ 
void game_raccoon_regenerate_directory_cache()
{   
   MCU_NEG68000_isolate_bus();
   ADDRESS_BUS_set_to_output();
   DATA_BUS_set_to_output();
   
   GAME_RACCOON_REGENERATE_DIRECTORY_CACHE_WORKSPACE w;

   /* record_index_write_cursor is the last written-to record index. */
   w.record_index_write_cursor    = 0;
   w.present_scanned_directory[0] = 0;
   strncat(w.present_scanned_directory, ".", DIRECTORY_CACHE_RECORD_DIRECTORY_FULLPATHNAME_FIELD_LENGTH);
   
   /**
    * Generate the incomplete root directory record and write it to Flash.
    * Since this is the first record, we need to erase the first sector.
    * Subsequent sectors will be erased as the record_index_write_cursor 
    * field is advanced over to the relevant boundaries.
    * The root record's contents begin at 1, and its parent is 0.
    */   
   LED_G_set_on();
   FLASH_erase_sector_at_address_in_words(DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(w.record_index_write_cursor));
   LED_G_set_off();
   LED_B_set_on();
   FLASH_game_raccoon_directory_cache_program_record_filename(w.record_index_write_cursor, w.present_scanned_directory);
   FLASH_game_raccoon_directory_cache_program_record_field(w.record_index_write_cursor, DIRECTORY_CACHE_RECORD_DIRECTORY_FIELD_RECORD_INDEX_PARENT,         0);
   LED_B_set_off();
   
   game_raccoon_regenerate_directory_cache_populate_record(0, &w);   
}

/**
 * We want to install the game whose filename resides in the directory
 * cache record index we received from REGISTEROE.
 */
void game_raccoon_install_game_to_game_partition_from_directory_record_index(uint16_t directory_record_index)
{
   uint32_t flash_address_in_words = DIRECTORY_CACHE_RECORD_FLASH_generate_address_in_words(directory_record_index);
   
   DIRECTORY_CACHE_RECORD record;
   FLASH_read(flash_address_in_words, &record, DIRECTORY_CACHE_RECORD_SIZE_BYTES);

   const char *filename_to_load = record.file_record.fullpathname;

   copy_sd_file_to_flash(filename_to_load, FLASH_PARTITION_GAME, 1);  
   
   /**
    * The last address used will be in the game partition,
    * so return the line to boot before we go back to the menu!
    */
   MCU_FLASH_GAME_NEGBOOT_set_to_boot();
}

/*
 * Entry point after reset.
 */
int main(int argc, char **argv)
{
   /* Do nothing until the PLL lock has been achieved. */
   while (!OSCCONbits.LOCK) {};
   
   /**
    * The default PIC state is all pins input, which allows the
    * control lines to be pulled appropriately by the resistors
    * for the Mega Drive to boot transparently into the Boot
    * partition.
    * However, we really want to ensure the transceivers are
    * very firmly in connect mode as soon as possible when the PIC
    * comes online.
    */
   MCU_NEG68000_connect_bus();
   
   /* Set the clock divisor to zero. */
   /* TODO: DOZE setting? */
   CLKDIV = 0b0000000000000000;
   
   /**
    * Clear all the reset status bits.
    * This also contains the SWDTEN software watchdog control, and
    * the Program Memory Power During Sleep bit.
    * Sections 9.2 and 24.2.5 say I should leave this set for
    * faster waking.
    */
   RCON &= 0b0000000100000000;
   
   /* Power down all the peripherals, except SPI1, SPI2 and Timers */
   PMD1 = 0b0000000011100001; /* Enable timers, SPI1, SPI2.  Disable I2C module 1, U2, U1, ADC1. (U1 and U2 are UARTs.) */
   PMD2 = 0b0001111100011111; /* Enable nothing. Disable Input Capture and Output Compare. */
   PMD3 = 0b0000011110000010; /* Enable nothing. Disable Comparator module, Real-Time Clock, Parallel Master Port, CRC check, I2C module 2. */
         
   /* Safely initialise the chip. We want everything to be inputs first! */      
   MCU_control_lines_set_directionality_and_default_state();
   timer2_setup_for_fatfs();
      
   f_mount(&fatfs, "", 0);
       
   /**
    * At this point the bus is CONNECTED. Do not attempt to
    * output any values on the address or data buses
    * or enable the Flash read, Flash write or Register read
    * control lines without first isolating the bus!!
    * 
    * The intended action is to use MCU_REGISTERCLK_idle_for_changes()
    * to enter an idle state watching REGISTERCLK, and when CLK occurs,
    * the bus will be isolated and processing can continue as required.
    * Processing terminates and control is passed back to the Mega Drive
    * when the bus is reconnected.
    * 
    * ENSURE that all buses are input before doing so!
    */
   
//   MCU_NEG68000_isolate_bus();
//   ADDRESS_BUS_set_to_output();
//   DATA_BUS_set_to_output();
//   FLASH_unlock_bypass_mode_begin();
//   game_raccoon_regenerate_directory_cache();
//   FLASH_read_to_file("directory_cache_contents.bin", GAME_RACCOON_DIRECTORY_CACHE_BASE_ADDRESS_WORDS, DIRECTORY_CACHE_RECORD_SIZE_BYTES * 100);
//   infinite_loop_pulse();
   
//#define DEBUG_WRITE_IMAGE "game_raccoon_boot_image_directory_cache.bin"
//#define DEBUG_WRITE_IMAGE_PARTITION FLASH_PARTITION_BOOT  
   
#ifdef DEBUG_WRITE_IMAGE    
#ifndef DEBUG_WRITE_IMAGE_PARTITION
#error "If you're going to debug write an image, you need to specify a partition number (0=boot, 1=game)"
#endif   
   
#ifdef DEBUG_WRITE_IMAGE_PARTITION   
   MCU_NEG68000_isolate_bus();
   ADDRESS_BUS_set_to_output();
   DATA_BUS_set_to_output();
   FLASH_unlock_bypass_mode_begin();
   FRESULT copy_success = copy_sd_file_to_flash(DEBUG_WRITE_IMAGE, DEBUG_WRITE_IMAGE_PARTITION, 1);
   
   infinite_loop_pulse();
#endif   
#endif   
   
   /* Pause until the GR handshake is received on !TIME. */
   game_raccoon_wait_for_handshake();
   
   /* We have the bus for a second, we can set up Flash bypass mode here and keep it on permanently. */
   DATA_BUS_set_to_output();
   FLASH_unlock_bypass_mode_begin();
   
   while (1)
   {
      /* Reconnect the bus and wait for future commands. */      
      uint16_t register_read_game_or_boot = MCU_REGISTERCLK_idle_for_changes();
      /* Bus is isolated, but all buses and data lines fully input at this point. */

      if (register_read_game_or_boot == REGISTERCLK_CONSTANT_BOOT_INSTALLED_GAME)
      {
         /**
          * We've received the 0xFFFF command to boot the installed game.
          * We need to disable the PIC entirely, make it inert and invisible.
          * Setting the NEGBOOT/GAME flag to game presents the game partition
          * to the Mega Drive.
          */
         game_raccoon_switch_to_game_mode_and_terminate();
      }
      else
      if (register_read_game_or_boot == REGISTERCLK_CONSTANT_REGENERATE_DIRECTORY_CACHE)
      {
         /**
          * We've received the 0xFFFE command to regenerate the directory
          * cache in the boot partition.
          */         
         game_raccoon_regenerate_directory_cache();
      }
      else
      {
         /* We want to install the game at slot register_read_game_or_boot. */
         game_raccoon_install_game_to_game_partition_from_directory_record_index(register_read_game_or_boot);
      }
   }
                  
   return 0;
}
